from IPython.core.getipython import get_ipython


class BaseModelRegression(object):

	def fit(self, X, y):
		return NotImplementedError

	def predict(self, X):
		return NotImplementedError

	def __visual_training__(self):
		return NotImplementedError

	def score(self, real, predicted):
		return NotImplementedError

	def sklearn_version(self):
		return NotImplementedError

	def __is_visual_on__(self):
		try:
			shell = get_ipython().__class__.__name__
			if shell == 'ZMQInteractiveShell':
				return True  # Jupyter notebook or qtconsole
			elif shell == 'TerminalInteractiveShell':
				return False  # Terminal running IPython
			elif shell == 'Shell':
				return True  # Google Colab
			else:
				return False  # Other type (?)
		except NameError:
			return False

	def __is_google_colab__(self):
		try:
			shell = get_ipython().__class__.__name__
			if shell == 'Shell':
				return True  # Google Colab
			else:
				return False  # Other type (?)
		except NameError:
			return False


class BaseModelClassification(object):

	def fit(self, X, y):
		return NotImplementedError

	def predict(self, X):
		return NotImplementedError

	def __visual_training__(self):
		return NotImplementedError

	def score(self, real, predicted):
		return NotImplementedError

	def sklearn_version(self):
		return NotImplementedError

	def __is_visual_on__(self):
		try:
			shell = get_ipython().__class__.__name__
			if shell == 'ZMQInteractiveShell':
				return True  # Jupyter notebook or qtconsole
			elif shell == 'TerminalInteractiveShell':
				return False  # Terminal running IPython
			elif shell == 'Shell':
				return True  # Google Colab
			else:
				return False  # Other type (?)
		except NameError:
			return False

	def __is_google_colab__(self):
		try:
			shell = get_ipython().__class__.__name__
			if shell == 'Shell':
				return True  # Google Colab
			else:
				return False  # Other type (?)
		except NameError:
			return False


class BaseModelClustering(object):

	def fit(self, X):
		return NotImplementedError

	def predict(self, X):
		return NotImplementedError

	def __visual_training__(self):
		return NotImplementedError

	def sklearn_version(self):
		return NotImplementedError

	def __is_visual_on__(self):
		try:
			shell = get_ipython().__class__.__name__
			if shell == 'ZMQInteractiveShell':
				return True  # Jupyter notebook or qtconsole
			elif shell == 'TerminalInteractiveShell':
				return False  # Terminal running IPython
			elif shell == 'Shell':
				return True  # Google Colab
			else:
				return False  # Other type (?)
		except NameError:
			return False

	def __is_google_colab__(self):
		try:
			shell = get_ipython().__class__.__name__
			if shell == 'Shell':
				return True  # Google Colab
			else:
				return False  # Other type (?)
		except NameError:
			return False


from ml_tutor.model import BaseModelRegression

import pandas as pd
import matplotlib.pyplot as plt

from sklearn.metrics import mean_squared_error


class DecisionTreeRegression(BaseModelRegression):

	def __init__(self, criterion='mse', max_depth=None, visual_training=True, feature_names=None):
		"""
		Defines Decision Tree Regressor model.

		:param criterion:
		:param max_depth: The maximum depth of the tree. If None, then nodes are expanded until all leaves are pure or until all leaves contain less than min_samples_split samples.
		:param visual_training: If True - the training process will be visualized [NOTE: only in Jupyter Notebook and Google Colab]
		:param feature_names: List of names for columns in a dataset. (List of strings)
		"""
		super(BaseModelRegression, self).__init__()

		# Number of classes in the dataset
		self.predictions = None

		# Dataset -> X = features | y = labels/classes
		self.X = None
		self.y = None

		# Number of neighbours used to make predictions
		self.criterion = criterion
		self.max_depth = max_depth

		# Visualization related parameters
		self.visual_training = visual_training
		if not super().__is_visual_on__():
			self.visual_training = False
			print("Visualization is only supported in Jupyter Notebook and Google Colab.")

		self.feature_names = feature_names

	def fit(self, X, y):
		"""
		Train the model using features (X) as training data and y as target values.

		:param X: Features from a dataset
		:param y: Target (classes) values (This is what you want to predict)
		"""
		self.X = X

		if isinstance(self.X, pd.DataFrame):
			self.feature_names = self.X.columns

		self.y = y

		from sklearn.tree import DecisionTreeRegressor

		self.classifier = DecisionTreeRegressor(criterion=self.criterion,
		                                        max_depth=self.max_depth)

		self.figure = self.classifier.fit(self.X, self.y)

		if self.visual_training:
			self.__visual_training__()

	def predict(self, X):
		"""
		This method performs predictions on the unseen data from your dataset.

		:param X: Data samples used to perform prediction on. (Generally a test set)
		:return: Predicted labels for each data sample
		"""
		return self.classifier.predict(X)

	def __visual_training__(self):
		"""
		Helper function used to crete real time visualization of the training process.
		"""
		# Import only relevant libraries for Jupyter Notebook if needed
		from IPython import display
		from sklearn import tree

		plt.figure(figsize=(30, 20))
		tree.plot_tree(self.figure, feature_names=self.feature_names, filled=True, rounded=True)
		display.display(plt.gcf())
		display.display()
		display.clear_output(wait=True)

	def score(self, real, predicted):
		"""
		Return the MSE computed on real vs. predicted classes.

		:param real: Expected targets(generally found in the dataset)
		:param predicted: Predicted values by the algorithm

		:return: Mean squared error computed on real vs. predicted classes [0. - 1.]
		"""
		assert len(real) == len(predicted)
		return mean_squared_error(real, predicted)

	def sklearn_version(self):
		"""
		Auto-generates sklearn code for a selected algorithm.

		NOTE: This function will automatically add one more code cell to your Jupyter Notebook/Google Colab (with the sklearn code inside).
		"""
		if not super().__is_visual_on__():
			print("Supported only in Jupyter Notebook and Google Colab.")
			return NotImplementedError

		if super().__is_google_colab__():
			return "This method is not supported in Google Colab for now :/"

		from IPython.core.getipython import get_ipython
		contents = """
# If you don't have Sklearn installed execute line below
# pip install sklearn

# This is how you can import DecisionTreeRegressor using sklearn library
from sklearn.tree import DecisionTreeRegressor

# Define classifier with selected parameters
model = DecisionTreeRegressor(criterion='mse', max_depth=None,)

# Train the model using dataset you desire
model.fit(X_train, y_train)

# Finally, use trained model to make predictions
predictions = model.predict(X_test)

# Use Score method to make predictions
print(model.score(X_test, y_test))
"""
		shell = get_ipython()
		payload = dict(
			source='set_next_input',
			text=contents,
			replace=False,
		)
		shell.payload_manager.write_payload(payload, single=False)

	def how_it_works(self, video=False):
		"""
		Generates theory on how the algorithm works right in the Jupyter Notebook/Google colab.

		:param video: Some people prefer video tutorials over reading version. Set this parameter to True if you want video tutorial instead. :)
		"""
		if not super().__is_visual_on__():
			print("Supported only in Jupyter Notebook and Google Colab.")
			return NotImplementedError

		from IPython.core.getipython import get_ipython
		if not video:
			content = u"""
<div>
<h1>Decision Tree Algorithm — Explained</h1>
<img src="https://miro.medium.com/max/440/0*r8DWyN5pX4DRU89g.gif">
<br>
<br>

<p>
<h2>Introduction</h2>
<br><br>Classification is a two-step process, learning step and prediction step, in machine learning. In the learning step, the model is developed based on given training data. In the prediction step, the model is used to predict the response for given data. Decision Tree is one of the easiest and popular classification algorithms to understand and interpret.
<br><br><br><br>
<h2>Decision Tree Algorithm</h2><br><br>
Decision Tree algorithm belongs to the family of supervised learning algorithms. Unlike other supervised learning algorithms, the decision tree algorithm can be used for solving regression and classification problems too.
<br><br>The goal of using a Decision Tree is to create a training model that can use to predict the class or value of the target variable by learning simple decision rules inferred from prior data(training data).
<br><br>In Decision Trees, for predicting a class label for a record we start from the root of the tree. We compare the values of the root attribute with the record’s attribute. On the basis of comparison, we follow the branch corresponding to that value and jump to the next node.

<br><br><br><br><h2>Types of Decision Trees</h2><br><br>
Types of decision trees are based on the type of target variable we have. It can be of two types:<br><br><br>
    Categorical Variable Decision Tree: Decision Tree which has a categorical target variable then it called a Categorical variable decision tree.
    <br><br>Continuous Variable Decision Tree: Decision Tree has a continuous target variable then it is called Continuous Variable Decision Tree.
<br><br>


Example:- Let’s say we have a problem to predict whether a customer will pay his renewal premium with an insurance company (yes/ no). Here we know that the income of customers is a significant variable but the insurance company does not have income details for all customers. Now, as we know this is an important variable, then we can build a decision tree to predict customer income based on occupation, product, and various other variables. In this case, we are predicting values for the continuous variables.


<br><br><br><br><h2>Important Terminology related to Decision Trees</h2><br><br><br>
    Root Node: It represents the entire population or sample and this further gets divided into two or more homogeneous sets.
    <br><br>Splitting: It is a process of dividing a node into two or more sub-nodes.
    <br><br>Decision Node: When a sub-node splits into further sub-nodes, then it is called the decision node.
    <br><br>Leaf / Terminal Node: Nodes do not split is called Leaf or Terminal node.
    <br><br>Pruning: When we remove sub-nodes of a decision node, this process is called pruning. You can say the opposite process of splitting.
    <br><br>Branch / Sub-Tree: A subsection of the entire tree is called branch or sub-tree.
    <br><br>Parent and Child Node: A node, which is divided into sub-nodes is called a parent node of sub-nodes whereas sub-nodes are the child of a parent node.
<br><br><br>
<img src="https://miro.medium.com/max/757/1*bcLAJfWN2GpVQNTVOCrrvw.png">
<br><br><br>Decision trees classify the examples by sorting them down the tree from the root to some leaf/terminal node, with the leaf/terminal node providing the classification of the example.
<br><br>Each node in the tree acts as a test case for some attribute, and each edge descending from the node corresponds to the possible answers to the test case. This process is recursive in nature and is repeated for every subtree rooted at the new node.


<br><br><br><br><h2>Assumptions while creating Decision Tree</h2><br><br><br>
    Below are some of the assumptions we make while using Decision tree:
    <br><br>In the beginning, the whole training set is considered as the root.
    <br><br>Feature values are preferred to be categorical. If the values are continuous then they are discretized prior to building the model.
    <br><br>Records are distributed recursively on the basis of attribute values.
    <br><br>Order to placing attributes as root or internal node of the tree is done by using some statistical approach.
<br><br><br>

Decision Trees follow Sum of Product (SOP) representation. The Sum of product (SOP) is also known as Disjunctive Normal Form. For a class, every branch from the root of the tree to a leaf node having the same class is conjunction (product) of values, different branches ending in that class form a disjunction (sum).
<br><br>
The primary challenge in the decision tree implementation is to identify which attributes do we need to consider as the root node and each level. Handling this is to know as the attributes selection. We have different attributes selection measures to identify the attribute which can be considered as the root note at each level.


<br><br><br><br><h2>How do Decision Trees work?</h2><br><br><br>
    The decision of making strategic splits heavily affects a tree’s accuracy. The decision criteria are different for classification and regression trees.
    <br><br>Decision trees use multiple algorithms to decide to split a node into two or more sub-nodes. The creation of sub-nodes increases the homogeneity of resultant sub-nodes. In other words, we can say that the purity of the node increases with respect to the target variable. The decision tree splits the nodes on all available variables and then selects the split which results in most homogeneous sub-nodes.
    <br><br>The algorithm selection is also based on the type of target variables. Let us look at some algorithms used in Decision Trees:
<br><br>ID3 → (extension of D3)
<br><br>C4.5 → (successor of ID3)
<br><br>CART → (Classification And Regression Tree)
<br><br>CHAID → (Chi-square automatic interaction detection Performs multi-level splits when computing classification trees)
<br><br>MARS → (multivariate adaptive regression splines)
<br><br><br>
The ID3 algorithm builds decision trees using a top-down greedy search approach through the space of possible branches with no backtracking. A greedy algorithm, as the name suggests, always makes the choice that seems to be the best at that moment.
<br><br><br><br>
Steps in ID3 algorithm:
<br><br>It begins with the original set S as the root node.
<br><br>On each iteration of the algorithm, it iterates through the very unused attribute of the set S and calculates Entropy(H) and Information gain(IG) of this attribute.
<br><br>It then selects the attribute which has the smallest Entropy or Largest Information gain.
<br><br>The set S is then split by the selected attribute to produce a subset of the data.
<br><br>The algorithm continues to recur on each subset, considering only attributes never selected before.

</p>



<h1>Author and source:</h1>
<h2>Author: <a target="_blank" href="https://twitter.com/nschauhan00">Nagesh Singh Chauhan</a></h2>
<h2>Read about overfitting and attributes in the source blog: <a target="_blank" href="https://towardsdatascience.com/decision-tree-algorithm-explained-83beb6e78ef4">Towards data science post</a></h2>

</div>
"""
			get_ipython().run_cell_magic(u'html', u'', content)
		else:
			content = u"""
<div>
<h1> Decision Tree - How it works? </h1>
<iframe width="560" height="315" src="https://www.youtube.com/embed/RmajweUFKvM" frameborder="0" allow="accelerometer; autoplay; encrypted-media; gyroscope; picture-in-picture" allowfullscreen></iframe>
</div>
"""
			get_ipython().run_cell_magic(u'markdown', u'', content)

	def interview_questions(self):
		"""
		Generates commonly asked interview questions about the algorithm in the Jupyter Notebook/Google colab.
		"""
		if not super().__is_visual_on__():
			print("Supported only in Jupyter Notebook and Google Colab.")
			return NotImplementedError

		from IPython.core.getipython import get_ipython

		content = u"""
<h1> Decision Tree Interview Questions </h1>

Quiz like questions: <a href="https://www.analyticsvidhya.com/blog/2017/09/30-questions-test-tree-based-models/" target="_blank">link</a>
"""
		get_ipython().run_cell_magic(u'html', u'', content)



from ml_tutor.model import BaseModelClassification

import random
import numpy as np
import matplotlib.pyplot as plt

from sklearn.decomposition import PCA


class LogisticRegression(BaseModelClassification):
	
	def __init__(self, learning_rate=0.1, num_iter=100000, fit_intercept=True, visual_training=True):
		"""
		Create Logistic Regression model.

		:param learning_rate: Learning rate gives the rate of speed where the gradient moves during gradient descent
		:param num_iter: Number of times to go through the dataset to train the model.
		:param fit_intercept: If True, bias will be used.
		:param visual_training: If True - the training process will be visualized [NOTE: only in Jupyter Notebook and Google Colab]
		"""
		super(BaseModelClassification, self).__init__()

		self.learning_rate = learning_rate
		self.num_iter = num_iter
		self.fit_intercept = fit_intercept
		
		self.visual_training = visual_training

		if not super().__is_visual_on__():
			self.visual_training = False
			print("Visualization is only supported in Jupyter Notebook and Google Colab.")

		self.pca = PCA(n_components=2)
		self.chars = '0123456789ABCDEF'
		self.colors = None
		
		self.theta = None
		self.theta_history = []
		self.threshold = 0.5
	
	def __add_intercept__(self, X):
		"""
		Add bias to the dataset
		:param X: Input data
		:return: Input data with concatenated column for bias
		"""
		return np.c_[np.ones((X.shape[0], 1)), X]
	
	def __sigmoid__(self, x):
		"""
		Sigmoid Activation function.
		:param x: Input data
		:return: Sigmoid values of the input data
		"""
		return 1 / (1 + np.exp(-x))

	def fit(self, X, y):
		"""
		Train the model using features (X) as training data and y as target values.

		:param X: Features from a dataset
		:param y: Target (classes) values (This is what you want to predict)
		"""
		from IPython import display
		
		self.X = X
		self.y = y
		
		if len(self.y.shape) < 2:
			self.y = np.expand_dims(self.y, axis=1)

		if len(self.X.shape) < 2:
			self.X = np.expand_dims(self.X, axis=1)
			
		self.num_of_classes = max(self.y)[0] + 1
		
		self.colors = ['#' + ''.join(random.sample(self.chars, 6)) for i in range(self.num_of_classes)]

		if self.num_of_classes > 2:
			message = "This algorithm supports ONLY binary classification. Max number of unique classes is two (2). \n Found: {} classes".format(self.num_of_classes)
			raise ValueError(message)
		
		if self.X.shape[1] > 2:
			if self.visual_training:
				print("The dataset is sparse for visual training. It will be pre-processed so it can be visualized.")
				print("Current shape of your data: {}".format(self.X.shape))
				self.X = self.pca.fit_transform(self.X)
				print("New shape of your data: {}".format(self.X.shape))
				
		if self.fit_intercept:
			self.X = self.__add_intercept__(self.X)
		
		# weights initialization
		self.theta = np.zeros((self.X.shape[1], 1))
		for i in range(self.num_iter):
			out = self.__sigmoid__(np.dot(self.X, self.theta))
			grad = np.dot(self.X.T, (out - self.y)) / self.y.size
			self.theta -= (self.learning_rate * grad)

			# NOTE: Visualization of the training process
			# TODO: Create better real time viz
			if self.visual_training:
				if i % 5000 == 0:
					# preds = self.visual_predict_prob(self.theta) >= self.threshold
			
					plt.close()
					plt.clf()
					plt.figure(figsize=(8, 6))

					for k in range(self.num_of_classes):
						plt.scatter(self.X[y == k, -2], self.X[y == k, -1], c=self.colors[k])

					x_values = [np.min(self.X[:, 1]) + 0.5, np.max(self.X[:, 2]) +1]
					y_values = - (self.theta[0] + np.dot(np.array(x_values).reshape(2, 1), self.theta[1].reshape(1, 1))) / self.theta[2]

					plt.plot(x_values, y_values, c='r', label="Classificatio line")
					display.display(plt.gcf())
					display.display()
					display.clear_output(wait=True)

	def predict_prob(self, X):
		"""
		This method performs predictions on the unseen data from your dataset but instead of return classes it returns probability of each class.

		:param X: Data samples used to perform prediction on. (Generally a test set)
		:return: Probability per class given data sample
		"""
		if self.fit_intercept:
			X = self.__add_intercept__(X)
	
		return self.__sigmoid__(np.dot(X, self.theta))
	
	def predict(self, X):
		"""
		This method performs predictions on the unseen data from your dataset.

		:param X: Data samples used to perform prediction on. (Generally a test set)
		:return: Predicted labels for each data sample
		"""
		return self.predict_prob(X) >= self.threshold

	def score(self, real, predicted):
		"""
		Return the accuracy computed on real vs. predicted classes.

		:param real: Expected targets(generally found in the dataset)
		:param predicted: Predicted classes by the algorithm

		:return: Mean accuracy computed on real vs. predicted classes [0. - 1.]
		"""
		assert len(real) == len(predicted)
		return sum(real == predicted) / len(real)

	def sklearn_version(self):
		"""
		Auto-generates sklearn code for a selected algorithm.

		NOTE: This function will automatically add one more code cell to your Jupyter Notebook/Google Colab (with the sklearn code inside).
		"""

		if not super().__is_visual_on__():
			print("Supported only in Jupyter Notebook and Google Colab.")
			return NotImplementedError

		if super().__is_google_colab__():
			return "This method is not supported in Google Colab for now :/"

		from IPython.core.getipython import get_ipython
		contents = """
# If you don't have Sklearn installed execute line below
# pip install sklearn

# This is how you can import DecisionTreeClassifier using sklearn library
from sklearn.linear_model import LogisticRegression

# Define classifier with selected parameters
model = LogisticRegression()

# Train the model using dataset you desire
model.fit(X_train, y_train)

# Finally, use trained model to make predictions
predictions = model.predict(X_test)

# Use Score method to make predictions
print(model.score(X_test, y_test))
"""
		shell = get_ipython()
		payload = dict(
			source='set_next_input',
			text=contents,
			replace=False,
		)
		shell.payload_manager.write_payload(payload, single=False)

	def how_it_works(self, video=False):
		"""
		Generates theory on how the algorithm works right in the Jupyter Notebook/Google colab.

		:param video: Some people prefer video tutorials over reading version. Set this parameter to True if you want video tutorial instead. :)
		"""

		if not super().__is_visual_on__():
			print("Supported only in Jupyter Notebook and Google Colab.")
			return NotImplementedError

		from IPython.core.getipython import get_ipython

		if not video:
			content = u"""
<div>
<h1>Logistic Regression Explained</h1>
<br>
<img src="https://miro.medium.com/max/770/1*QY3CSyA4BzAU6sEPFwp9ZQ.png">
<br>
<br>
<p>

In the Machine Learning world, Logistic Regression is a kind of parametric classification model, despite having the word ‘regression’ in its name.<br><br>
This means that logistic regression models are models that have a certain fixed number of parameters that depend on the number of input features, and they output categorical prediction, like for example if a plant belongs to a certain species or not.

In Logistic Regression, we don’t directly fit a straight line to our data like in linear regression. Instead, we fit a S shaped curve, called Sigmoid, to our observations.
</p>
<br><br>
<img src="https://miro.medium.com/max/770/1*44qV8LhNzE5hPnta2PaaHw.png">
<br><br>
<p>
Let's examine this figure closely.
<br><br>First of all, like we said before, Logistic Regression models are classification models; specifically binary classification models (they can only be used to distinguish between 2 different categories — like if a person is obese or not given its weight, or if a house is big or small given its size). This means that our data has two kinds of observations (Category 1 and Category 2 observations) like we can observe in the figure.
<br><br>Note: This is a very simple example of Logistic Regression, in practice much harder problems can be solved using these models, using a wide range of features and not just a single one.
<br><br>Secondly, as we can see, the Y-axis goes from 0 to 1. This is because the sigmoid function always takes as maximum and minimum these two values, and this fits very well our goal of classifying samples in two different categories. By computing the sigmoid function of X (that is a weighted sum of the input features, just like in Linear Regression), we get a probability (between 0 and 1 obviously) of an observation belonging to one of the two categories.
<br><br>The formula for the sigmoid function is the following:
<br><br>
<img src="https://miro.medium.com/max/316/0*59BSXTBcxZcZGtVT">
<br><br>
<h2>1) Calculate weighted sum of inputs</h2>
<img src="https://miro.medium.com/max/271/0*vq7V-FuK9EirWDeN">
<br><br>
<h2>2) Calculate the probability of Obese</h2>
<img src="https://miro.medium.com/max/494/0*p5Yczl6itusXkxN8">
<br><br>
Alright, this looks cool and all, but isn’t this meant to be a Machine Learning model? How do we train it? That is a good question. There are multiple ways to train a Logistic Regression model (fit the S shaped line to our data). We can use an iterative optimisation algorithm like Gradient Descent to calculate the parameters of the model (the weights) or we can use probabilistic methods like Maximum likelihood.
</p>
<p>

<br><br>
Once we have used one of these methods to train our model, we are ready to make some predictions. Let's see an example of how the process of training a Logistic Regression model and using it to make predictions would go:
<br><br>1. First, we would collect a Dataset of patients who have and who have not been diagnosed as obese, along with their corresponding weights.

<br><br>2. After this, we would train our model, to fit our S shape line to the data and obtain the parameters of the model. After training using Maximum Likelihood, we got the following parameters:

<br><br>
<img src="https://miro.medium.com/max/487/1*sGI7P3PLzVcwWkLTE5Q7mg.png">

<br><br>
3. Now, we are ready to make some predictions: imagine we got two patients; one is 120 kg and one is 60 kg. Let's see what happens when we plug these numbers into the model:

<br><br>
<img src="https://miro.medium.com/max/520/1*fCwfXTnE55D_MJWNqhkegQ.png">


<br><br>
<img src="https://miro.medium.com/max/483/1*dwEPZB0Y6wKBvVNoQTLy6A.png">
</p>

<h1>Author and source:</h1>
<h2>Author: <a href="https://twitter.com/Jaimezorno">Jaime Zornoza</a></h2>
<h2>To find more resources go to the source post: <a href="https://towardsdatascience.com/logistic-regression-explained-9ee73cede081">Towards data science post</a></h2>

</div>
"""
			get_ipython().run_cell_magic(u'html', u'', content)
		else:
			content = u"""
<div>
<h1> Logistic Regression - How it works? </h1>
<iframe width="560" height="315" src="https://www.youtube.com/embed/-la3q9d7AKQ" frameborder="0" allow="accelerometer; autoplay; encrypted-media; gyroscope; picture-in-picture" allowfullscreen></iframe>
</div>
"""
			get_ipython().run_cell_magic(u'html', u'', content)

	def interview_questions(self):
		"""
		Generates commonly asked interview questions about the algorithm in the Jupyter Notebook/Google colab.
		"""
		if not super().__is_visual_on__():
			print("Supported only in Jupyter Notebook and Google Colab.")
			return NotImplementedError

		from IPython.core.getipython import get_ipython

		content = u"""
<h1> Logistic Regression Interview Questions </h1>

Quiz like questions: <a href="https://www.analyticsvidhya.com/blog/2017/08/skilltest-logistic-regression/" target="_blank">link</a>
"""
		get_ipython().run_cell_magic(u'html', u'', content)




from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()


install_requires = ["matplotlib>=3.2.1",
                    "pandas>=0.25.1",
                    "numpy>=1.18.1",
                    "scikit_learn>=0.23.2"
                    ]

setup(name='ml_tutor',
      version='1.0.0.1',
      description='ML Tutor : Learn Machine Learning while never leaving the conform of your Python IDE (Jupyter Notebook or Google Colab)',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/lucko515/ml_tutor',
      author='Luka Anicin',
      author_email='luka.anicin@gmail.com',
      license='MIT',
      install_requires=install_requires,
      packages=find_packages(),
      classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            "Intended Audience :: Education",
            "Intended Audience :: Developers",
            "Intended Audience :: Information Technology",
            "Intended Audience :: Science/Research"
      ],
      python_requires='>=3',
      )



from ml_tutor.model import BaseModelRegression

import time
import numpy as np
import matplotlib.pyplot as plt
from sklearn.metrics import mean_squared_error


class LinearRegression(BaseModelRegression):

	def __init__(self, learning_rate=0.0001, num_iter=100000, tol=0.00001, visual_training=True):
		"""
		Creates the Linear Regression algorithm.

		:param learning_rate: Learning rate gives the rate of speed where the gradient moves during gradient descent
		:param num_iter: Number of times to go through the dataset to train the model.
		:param tol: If the difference between old and new values for model parameters are less than this number, training stops.
		:param visual_training: If True - the training process will be visualized [NOTE: only in Jupyter Notebook and Google Colab]
		"""
		super(BaseModelRegression, self).__init__()

		self.learning_rate = learning_rate
		self.num_iter = num_iter
		self.tol = tol
		self.visual_training = visual_training
		if not super().__is_visual_on__():
			self.visual_training = False
			print("Visualization is only supported in Jupyter Notebook and Google Colab.")

		self.randn_id = None

		# Gradient descent params
		self.starting_b = 0
		self.starting_m = 0
		self.b_history = []
		self.m_history = []

		print("If your dataset is sparse for visual training, random feature will be selected to match required shape.")
		print("Required shape for this algorithm is: [N, 1].")

	def fit(self, X, y):
		"""
		Train the model using features (X) as training data and y as target values.

		:param X: Features from a dataset
		:param y: Target values (This is what you want to predict)
		"""

		self.X = X
		self.y = y

		if len(self.y.shape) < 2:
			self.y = np.expand_dims(self.y, axis=1)

		if len(self.X.shape) < 2:
			self.X = np.expand_dims(self.X, axis=1)

		if self.X.shape[1] > 1:
			if self.visual_training:
				print("The dataset is sparse for visual training. This algorithm works only on shape [N, 1].")
				print("Random feature selected to match required size.")
				print("Current shape of your data: {}".format(self.X.shape))
				self.randn_id = np.random.randint(0, self.X.shape[1])
				print("Column selected on id: {}".format(self.randn_id))
				self.X = self.X[:, self.randn_id]
				if len(self.X.shape) < 2:
					self.X = np.expand_dims(self.X, axis=1)
				print("New shape of your data: {}".format(self.X.shape))

		# calling gradient descent function, and output of it is going to be our the best possible (according to our dataset) M and B
		self.__gradient_descent__(self.starting_b, self.starting_m)

	def __gradient_descent__(self, b, m):
		"""
		 main function for the gradient descent
		 :param b: Bias or constant
		 :param m: coefficient for X
		"""
		self.new_b = b
		self.new_m = m

		for i in range(self.num_iter):
			candidate_m, candidate_b = self.__gradient_descent_step__(self.new_b, self.new_m)

			if all(np.abs(candidate_m - self.new_m) <= self.tol) and \
					all(np.abs(candidate_b - self.new_b) <= self.tol):
				break

			self.new_m = candidate_m
			self.new_b = candidate_b

			if i % 1000 == 0:
				self.b_history.append(self.new_b)
				self.m_history.append(self.new_m)

		if self.visual_training:
			self.__visual_training__()

	def __visual_training__(self):
		"""
		Helper function used to crete real time visualization of the training process.
		"""

		# Import only relevant libraries for Jupyter Notebook if needed
		from IPython import display

		for i in range(len(self.b_history)):
			plt.close()
			plt.clf()
			plt.figure(figsize=(12, 10))

			plt.scatter(self.X, self.y, c='b', label="Training set")
			plt.plot(self.X, np.add(np.multiply(self.X, self.m_history[i]), self.b_history[i]), c='r',
			         label="Regression line")
			plt.title("Linear Regression - Training process")
			plt.xlabel("Feature value")
			plt.ylabel("Target value")
			plt.legend(framealpha=1, frameon=True)

			display.display(plt.gcf())
			display.display()
			time.sleep(1)
			display.clear_output(wait=True)

	def __gradient_descent_step__(self, b, m):
		"""
		Helper function for Gradient descent. Performs a single step of the gradient optimization.
		"""
		candidated_b = b - np.multiply(self.learning_rate,
		                               np.sum(-np.multiply(2 / float(len(self.X)),
		                                                   np.subtract(self.y,
		                                                               np.add(np.multiply(self.X, m), b))), axis=0))

		candidated_m = m - np.multiply(self.learning_rate,
		                               np.sum(np.multiply(2 / float(len(self.X)),
		                                                  np.multiply(-self.X,
		                                                              np.subtract(self.y,
		                                                                          np.add(np.multiply(self.X, m), b)))),
		                                      axis=0))

		return candidated_m, candidated_b

	def predict(self, X):
		"""
		This method performs predictions on the unseen data from your dataset.

		:param X: Data samples used to perform prediction on. (Generally a test set)
		:return: Predicted labels for each data sample
		"""
		if X.shape[1] > 2:
			if self.visual_training:
				X = X[:, self.randn_id]

		if X.shape[1] < 2:
			X = np.expand_dims(X, axis=1)

		y_pred = np.add(np.multiply(X, self.new_m), self.new_b)

		return y_pred

	def score(self, real, predicted):
		"""
		Return the MSE computed on real vs. predicted classes.

		:param real: Expected targets(generally found in the dataset)
		:param predicted: Predicted values by the algorithm

		:return: Mean squared error computed on real vs. predicted classes [0. - 1.]
		"""
		assert len(real) == len(predicted)
		return mean_squared_error(real, predicted)

	def sklearn_version(self):
		"""
		Auto-generates sklearn code for a selected algorithm.

		NOTE: This function will automatically add one more code cell to your Jupyter Notebook/Google Colab (with the sklearn code inside).
		"""
		if not super().__is_visual_on__():
			print("Supported only in Jupyter Notebook and Google Colab.")
			return NotImplementedError

		if super().__is_google_colab__():
			return "This method is not supported in Google Colab for now :/"

		from IPython.core.getipython import get_ipython
		contents = """
# If you don't have Sklearn installed execute line below
# pip install sklearn

# This is how you can import LinearRegression using sklearn library
from sklearn.linear_model import LinearRegression

# Define regressor with selected parameters
model = LinearRegression()

# Train the model using dataset you desire
model.fit(X_train, y_train)

# Finally, use trained model to make predictions
predictions = model.predict(X_test)

# Use Score method to make predictions
print(model.score(X_test, y_test))
"""
		shell = get_ipython()
		payload = dict(
			source='set_next_input',
			text=contents,
			replace=False,
		)
		shell.payload_manager.write_payload(payload, single=False)

	def how_it_works(self, video=False):
		"""
		Generates theory on how the algorithm works right in the Jupyter Notebook/Google colab.

		:param video: Some people prefer video tutorials over reading version. Set this parameter to True if you want video tutorial instead. :)
		"""
		if not super().__is_visual_on__():
			print("Supported only in Jupyter Notebook and Google Colab.")
			return NotImplementedError

		from IPython.core.getipython import get_ipython

		if not video:
			content = u"""
<div>
<h1>Linear Regression — Understanding the Theory</h1>
<br>
<img src="https://miro.medium.com/max/770/0*c39Seo5WzCpU4GAn">
<br>
<br>
<p>

Linear regression is probably the simplest approach for statistical learning. It is a good starting point for more advanced approaches, and in fact, many fancy statistical learning techniques can be seen as an extension of linear regression. Therefore, understanding this simple model will build a good base before moving on to more complex approaches.
<br><br>
Linear regression is very good to answer the following questions:<br><br>
- Is there a relationship between 2 variables?<br>
- How strong is the relationship?<br>
- Which variable contributes the most?<br>
- How accurately can we estimate the effect of each variable?<br>
- How accurately can we predict the target?<br>
- Is the relationship linear? (duh)<br>
- Is there an interaction effect?<br>
</p>

<p>

<h2>Estimating the coefficients</h2><br><br>
Let’s assume we only have one variable and one target. Then, linear regression is expressed as:
<br><br>
<img src="https://miro.medium.com/max/770/1*B-U6j1vxqqaYjgTZgunxIg@2x.png">

<br><br>
In the equation above, the betas are the coefficients. These coefficients are what we need in order to make predictions with our model.<br><br>
So how do we find these parameters?<br><br>
To find the parameters, we need to minimize the least squares or the sum of squared errors. Of course, the linear model is not perfect and it will not predict all the data accurately, meaning that there is a difference between the actual value and the prediction. The error is easily calculated with:
<br><br>
<img src="https://miro.medium.com/max/727/1*ly-QBw2oLDVx9M7MzxkKnw@2x.png">
<br><br>
But why are the errors squared?<br><br>
We square the error, because the prediction can be either above or below the true value, resulting in a negative or positive difference respectively. If we did not square the errors, the sum of errors could decrease because of negative differences and not because the model is a good fit. Also, squaring the errors penalizes large differences, and so the minimizing the squared errors “guarantees” a better model. Let’s take a look at a graph to better understand.
<br><br>
<img src="https://miro.medium.com/max/770/1*3CgiH8QI0ZN5LfdmK2t6XQ@2x.png">
<br><br>
In the graph above, the red dots are the true data and the blue line is linear model. The grey lines illustrate the errors between the predicted and the true values. The blue line is thus the one that minimizes the sum of the squared length of the grey lines.
<br><br>After some math that is too heavy for a blog post, you can finally estimate the coefficients with the following equations:<br><br>
<br><br>
<img src="https://miro.medium.com/max/614/1*YOiQ9UpR-A2jHvGR6JZwtQ@2x.png"><br><br>
<img src="https://miro.medium.com/max/339/1*t9rzyx0zh7o5Zx1Y-IQOvg@2x.png">
<br><br>
Where x bar and y bar represent the mean.
<br><br>

<h2>Estimate the relevancy of the coefficients</h2>
<br><br>

Now that you have coefficients, how can you tell if they are relevant to predict your target?<br><br>
The best way is to find the p-value. The p-value is used to quantify statistical significance; it allows to tell whether the null hypothesis is to be rejected or not.<br><br>
The null hypothesis?<br><br>
For any modelling task, the hypothesis is that there is some correlation between the features and the target. The null hypothesis is therefore the opposite: there is no correlation between the features and the target.<br><br>
So, finding the p-value for each coefficient will tell if the variable is statistically significant to predict the target. As a general rule of thumb, if the p-value is less than 0.05: there is a strong relationship between the variable and the target.
<br><br>
<h2>Assess the accuracy of the model</h2>
<br><br>

You found out that your variable was statistically significant by finding its p-value. Great!<br><br>
Now, how do you know if your linear model is any good?<br><br>
To assess that, we usually use the RSE (residual standard error) and the R² statistic.<br><br>

The first error metric is simple to understand: the lower the residual errors, the better the model fits the data (in this case, the closer the data is to a linear relationship).<br><br>
As for the R² metric, it measures the proportion of variability in the target that can be explained using a feature X. Therefore, assuming a linear relationship, if feature X can explain (predict) the target, then the proportion is high and the R² value will be close to 1. If the opposite is true, the R² value is then closer to 0.
<br><br>
</p>



<h1>Author and source:</h1>
<h2>Author: <a target="_blank" href="https://towardsdatascience.com/@marcopeixeiro">Marco Peixeiro</a></h2>
<h2>To find more resources go to the source of the post: <a target="_blank" href="https://towardsdatascience.com/linear-regression-understanding-the-theory-7e53ac2831b5">Towards data science post</a></h2>

</div>
"""
			get_ipython().run_cell_magic(u'html', u'', content)
		else:
			content = u"""
<div>
<h1> K-Means - How it works? </h1>
<iframe width="560" height="315" src="https://www.youtube.com/embed/kHwlB_j7Hkc" frameborder="0" allow="accelerometer; autoplay; encrypted-media; gyroscope; picture-in-picture" allowfullscreen></iframe>
</div>
"""
			get_ipython().run_cell_magic(u'markdown', u'', content)

	def interview_questions(self):
		"""
		Generates commonly asked interview questions about the algorithm in the Jupyter Notebook/Google colab.
		"""
		if not super().__is_visual_on__():
			print("Supported only in Jupyter Notebook and Google Colab.")
			return NotImplementedError

		from IPython.core.getipython import get_ipython

		content = u"""
<h1> Linear Regression Interview Questions </h1>

Quiz like questions: <a href="https://www.analyticsvidhya.com/blog/2017/07/30-questions-to-test-a-data-scientist-on-linear-regression/" target="_blank">link</a>	
"""
		get_ipython().run_cell_magic(u'html', u'', content)




from ml_tutor.model import BaseModelClustering

import time
import random
import numpy as np
import matplotlib.pyplot as plt
from sklearn.decomposition import PCA


class KMeans(BaseModelClustering):

	def __init__(self, n_clusters=2, tolerance=0.001, max_iter=300, visual_training=True):
		"""

		:param n_clusters:
		:param tolerance:
		:param max_iter:
		:param visual_training:
		"""
		super(BaseModelClustering, self).__init__()

		self.k_clusters = n_clusters
		self.tol = tolerance
		self.max_iter = max_iter

		# Visualization related parameters
		self.pca = PCA(n_components=2)
		self.visual_training = visual_training
		if not super().__is_visual_on__():
			self.visual_training = False
			print("Visualization is only supported in Jupyter Notebook and Google Colab.")

		# These two lists will be used to store mid steps for centroids and data, so we can visualize each step
		self.clustered_data_history = []
		self.centroids_history = []

	def fit(self, X):
		"""
		Train the model using features (X) as training data.

		:param X: Features from a dataset
		"""

		self.X = X

		if self.X.shape[1] > 2:
			if self.visual_training:
				print("The dataset is sparse for visual training. It will be pre-processed so it can be visualized.")
				print("Current shape of your data: {}".format(self.X.shape))
				self.X = self.pca.fit_transform(self.X)
				print("New shape of your data: {}".format(self.X.shape))

		# Starting clusters will be random members from X set
		self.centroids = []

		for i in range(self.k_clusters):
			# this index is used to acces random element from input set
			index = random.randint(1, len(self.X) - 1)
			self.centroids.append(self.X[index])

		for i in range(self.max_iter):
			# storing previous values of centroids
			prev_centroids = self.centroids
			# This will be dict for plotting data later on
			# with it we can find data points which are in the some cluster

			self.clustered_data = {}
			# Centroids values for this iteration
			cen_temp = []

			diffs = []
			for centroid in self.centroids:
				diffs.append(np.linalg.norm(self.X - centroid, axis=1))

			diffs = np.argmin(np.vstack(diffs).T, axis=1)

			for c in range(len(self.centroids)):
				closer_samples = self.X[diffs == c]
				self.clustered_data[c] = closer_samples
				cen_temp.append(np.average(closer_samples, axis=0))

			# Saving data and centroids for visual training
			if self.visual_training:
				self.clustered_data_history.append(self.clustered_data)
				self.centroids_history.append(self.centroids)

			# Check if it is optimized
			optimized = True
			for c in range(len(self.centroids)):
				original_centroid = prev_centroids[c]
				current_centroid = cen_temp[c]
				if np.abs(np.sum((current_centroid - original_centroid) / original_centroid * 100.0)) > self.tol:
					optimized = False
					self.centroids = cen_temp

			if optimized:
				break

		if self.visual_training:
			self.__visual_training__()

	def __visual_training__(self):
		"""
		Helper function used to crete real time visualization of the training process.
		"""
		# Import only relevant libraries for Jupyter Notebook if needed
		from IPython import display

		for i in range(len(self.centroids_history)):

			data = self.clustered_data_history[i]
			centroids = self.centroids_history[i]

			plt.close()
			plt.clf()
			plt.figure(figsize=(12, 10))

			for k in range(self.k_clusters):
				plt.scatter(np.vstack(data[k])[:, 0], np.vstack(data[k])[:, 1], label="Cluster {} data samples".format(k))

			centroid_vis = False
			for p in range(self.k_clusters):
				if not centroid_vis:
					plt.scatter(centroids[p][0], centroids[p][1], s=150,
					            c='black', marker="s", label="Centroid")
					centroid_vis = True
				else:
					plt.scatter(centroids[p][0], centroids[p][1], s=150,
					            c='black', marker="s")

			plt.title("K-Means - Training process")
			plt.legend(framealpha=1, frameon=True)

			display.display(plt.gcf())
			display.display()
			time.sleep(1)
			display.clear_output(wait=True)

	def predict(self, X):
		"""
		This method performs predictions on the unseen data from your dataset.

		:param X: Data samples used to perform prediction on. (Generally a test set)
		:return: Predicted labels for each data sample
		"""

		if X.shape[1] > 2:
			if self.visual_training:
				X = self.pca.transform(X)

		diffs = []
		for centroid in self.centroids:
			diffs.append(np.linalg.norm(X - centroid, axis=1))

		predictions = np.argmin(np.vstack(diffs).T, axis=1)

		return predictions

	def sklearn_version(self):
		"""
		Auto-generates sklearn code for a selected algorithm.

		NOTE: This function will automatically add one more code cell to your Jupyter Notebook/Google Colab (with the sklearn code inside).
		"""
		if not super().__is_visual_on__():
			print("Supported only in Jupyter Notebook and Google Colab.")
			return NotImplementedError
		if super().__is_google_colab__():
			return "This method is not supported in Google Colab for now :/"

		from IPython.core.getipython import get_ipython
		contents = """
# If you don't have Sklearn installed execute line below
# pip install sklearn

# This is how you can import KMeans using sklearn library
from sklearn.cluster import KMeans

# Define classifier with selected parameters
model = KMeans(n_clusters=3)

# Train the model using dataset you desire
model.fit(X_train)

# Finally, use trained model to make predictions
predictions = model.predict(X_test)
"""
		shell = get_ipython()
		payload = dict(
			source='set_next_input',
			text=contents,
			replace=False,
		)
		shell.payload_manager.write_payload(payload, single=False)

	def how_it_works(self, video=False):
		"""
		Generates theory on how the algorithm works right in the Jupyter Notebook/Google colab.

		:param video: Some people prefer video tutorials over reading version. Set this parameter to True if you want video tutorial instead. :)
		"""
		if not super().__is_visual_on__():
			print("Supported only in Jupyter Notebook and Google Colab.")
			return NotImplementedError

		from IPython.core.getipython import get_ipython
		if not video:
			content = u"""
<div>
<h1>Understanding K-means Clustering in Machine Learning</h1>
<br>
<br>

<p>

<br><br>K-means clustering is one of the simplest and popular unsupervised machine learning algorithms.
<br><br>Typically, unsupervised algorithms make inferences from datasets using only input vectors without referring to known, or labelled, outcomes.
<br><br>A cluster refers to a collection of data points aggregated together because of certain similarities.
<br><br>You’ll define a target number k, which refers to the number of centroids you need in the dataset. A centroid is the imaginary or real location representing the center of the cluster.
<br><br>Every data point is allocated to each of the clusters through reducing the in-cluster sum of squares.
<br><br>In other words, the K-means algorithm identifies k number of centroids, and then allocates every data point to the nearest cluster, while keeping the centroids as small as possible.
<br><br>The ‘means’ in the K-means refers to averaging of the data; that is, finding the centroid.

<br><br><br><br><h2>How the K-means algorithm works</h2><br><br>
<br><br>To process the learning data, the K-means algorithm in data mining starts with a first group of randomly selected centroids, which are used as the beginning points for every cluster, and then performs iterative (repetitive) calculations to optimize the positions of the centroids
<br><br>It halts creating and optimizing clusters when either:<br><br><br><br>
    The centroids have stabilized — there is no change in their values because the clustering has been successful.
    <br><br>The defined number of iterations has been achieved.
<br><br>
</p>



<h1>Author and source:</h1>
<h2>Author: <a target="_blank" href="https://towardsdatascience.com/@ledumjg">Dr. Michael J. Garbade</a></h2>
<h2>To find more resources go to the source of the post: <a target="_blank" href="https://towardsdatascience.com/understanding-k-means-clustering-in-machine-learning-6a6e67336aa1">Towards data science post</a></h2>

</div>
"""
			get_ipython().run_cell_magic(u'html', u'', content)
		else:
			content = u"""
<div>
<h1> K-Means - How it works? </h1>
<iframe width="560" height="315" src="https://www.youtube.com/embed/ZueoXMgCd1c" frameborder="0" allow="accelerometer; autoplay; encrypted-media; gyroscope; picture-in-picture" allowfullscreen></iframe>
</div>
"""
			get_ipython().run_cell_magic(u'markdown', u'', content)

	def interview_questions(self):
		"""
		Generates commonly asked interview questions about the algorithm in the Jupyter Notebook/Google colab.
		"""
		if not super().__is_visual_on__():
			print("Supported only in Jupyter Notebook and Google Colab.")
			return NotImplementedError

		from IPython.core.getipython import get_ipython

		content = u"""
<h1> K-Means Interview Questions </h1>

Quiz like questions: <a href="https://www.analyticsvidhya.com/blog/2017/02/test-data-scientist-clustering/" target="_blank">link</a>	
"""
		get_ipython().run_cell_magic(u'html', u'', content)




from ml_tutor.model import BaseModelClassification

from sklearn.decomposition import PCA
import random
import numpy as np
import matplotlib.pyplot as plt


class KNeighbourClassifier(BaseModelClassification):

	def __init__(self, n_neighbors=5, visual_training=True, number_of_visual_steps=-1):
		"""
		Create the K-Nearest Neighbors algorithm.

		:param n_neighbors: Number of neighbors to use when predicting a new data sample.
		:param visual_training: If True - the training process will be visualized [NOTE: only in Jupyter Notebook and Google Colab]
		:param number_of_visual_steps: Visualization might take a long time for this algorithm.
				With this parameter you can set how many points in the predictions time to visualize.
				NOTE: If you set this to -1 (by default) it will visualize ALL data samples you provide to it.
		"""
		super(BaseModelClassification, self).__init__()

		# Number of classes in the dataset
		# PCA is an unsupervised machine learning algorithm used to compress dataset and prepare it for visualization
		self.pca = PCA(n_components=2)
		self.predictions = None
		self.X_predict = None
		self.num_of_classes = None

		# Dataset -> X = features | y = labels/classes
		self.X = None
		self.y = None

		# Number of neighbours used to make predictions
		self.k = n_neighbors

		# Visualization related parameters
		self.visual_training = visual_training

		if not super().__is_visual_on__():
			self.visual_training = False
			print("Visualization is only supported in Jupyter Notebook and Google Colab.")

		self.number_of_visual_steps = number_of_visual_steps
		self.chars = '0123456789ABCDEF'
		self.colors = None

		# List that stores all distances
		self.distances = None

	def fit(self, X, y):
		"""
		Train the model using features (X) as training data and y as target values.

		:param X: Features from a dataset
		:param y: Target (classes) values (This is what you want to predict)
		"""
		self.X = X
		self.y = y

		if self.X.shape[1] > 2:
			if self.visual_training:
				print("The dataset is sparse for visual training. It will be pre-processed so it can be visualized.")
				print("Current shape of your data: {}".format(self.X.shape))
				self.X = self.pca.fit_transform(self.X)
				print("New shape of your data: {}".format(self.X.shape))

		self.num_of_classes = max(self.y) + 1
		self.colors = ['#' + ''.join(random.sample(self.chars, 6)) for i in range(self.num_of_classes)]

	def predict(self, X):
		"""
		This method performs predictions on the unseen data from your dataset.

		:param X: Data samples used to perform prediction on. (Generally a test set)
		:return: Predicted labels for each data sample
		"""

		if X.shape[1] > 2:
			if self.visual_training:
				X = self.pca.transform(X)

		self.X_predict = X
		self.predictions = []
		self.distances = []

		for i in range(len(X)):
			point = X[i]
			dist = np.linalg.norm(self.X - point, axis=1)
			self.distances.append(dist)
			idx = np.argpartition(dist, self.k)[:self.k]

			pred = np.argmax(np.bincount(self.y[idx]))
			self.predictions.append(pred)

		if self.visual_training:
			self.__visual_training__()

		return self.predictions

	def __visual_training__(self):
		"""
		Helper function used to crete real time visualization of the training process.
		"""
		# Import only relevant libraries for Jupyter Notebook if needed
		from IPython import display

		X = self.X_predict

		if self.number_of_visual_steps == -1:
			self.number_of_visual_steps = len(X)

		predicted_points = []

		for i in range(len(X)):
			point = X[i]
			dist = self.distances[i]
			if i < self.number_of_visual_steps:
				plt.close()
				plt.clf()
				plt.figure(figsize=(8, 6))
				plt.scatter(point[0], point[1], s=100)

				for k in range(self.num_of_classes):
					plt.scatter(self.X[self.y == k, 0], self.X[self.y == k, 1], c=self.colors[k], label="Class {} data samples".format(k))

				# Helper list for viz
				predicted_classes = []
				for p in range(len(predicted_points)):
					if self.predictions[p] not in predicted_classes:
						predicted_classes.append(self.predictions[p])
						plt.scatter(predicted_points[p][0], predicted_points[p][1], s=100,
						            c=self.colors[self.predictions[p]], label="Predicted points")
					else:
						plt.scatter(predicted_points[p][0], predicted_points[p][1], s=100,
						            c=self.colors[self.predictions[p]])

				for t in range(len(self.X)):
					current_x = self.X[t]

					rest_ids = []
					for j in range(len(self.X)):
						if j != i:
							rest_ids.append(j)

					distance_line = plt.plot([point[0], current_x[0]], [point[1], current_x[1]], "b", label="Distance line")

					m = [(point[k] + current_x[k]) / 2. for k in (0, 1)]
					text = plt.text(m[0], m[1], "%.2f" % dist[i])
					star = plt.scatter(current_x[0], current_x[1], c='black', marker="*")
					plt.title("KNeighbourClassifier - Prediction process")

					plt.legend(framealpha=1, frameon=True)
					display.display(plt.gcf())
					display.display()
					distance_line[0].remove()
					text.remove()
					star.remove()
					display.clear_output(wait=True)

				predicted_points.append(point)

	def score(self, real, predicted):
		"""
		Return the accuracy computed on real vs. predicted classes.

		:param real: Expected targets(generally found in the dataset)
		:param predicted: Predicted classes by the algorithm

		:return: Mean accuracy computed on real vs. predicted classes [0. - 1.]
		"""
		assert len(real) == len(predicted)
		return sum(real == predicted) / len(real)

	def sklearn_version(self):
		"""
		Auto-generates sklearn code for a selected algorithm.

		NOTE: This function will automatically add one more code cell to your Jupyter Notebook/Google Colab (with the sklearn code inside).
		"""
		if not super().__is_visual_on__():
			print("Supported only in Jupyter Notebook and Google Colab.")
			return NotImplementedError


		from IPython.core.getipython import get_ipython

		if super().__is_google_colab__():
			return "This method is not supported in Google Colab for now :/"

		contents = """
# If you don't have Sklearn installed execute line below
# pip install sklearn

# This is how you can import KNN using sklearn library
from sklearn.neighbors import KNeighborsClassifier

# Define classifier with selected parameters
model = KNeighborsClassifier(n_neighbors=5)

# Train the model using dataset you desire
model.fit(X_train, y_train)

# Finally, use trained model to make predictions
predictions = model.predict(X_test)

# Use Score method to make predictions
print(model.score(X_test, y_test))
"""
		shell = get_ipython()
		payload = dict(
			source='set_next_input',
			text=contents,
			replace=False,
		)
		shell.payload_manager.write_payload(payload, single=False)

	def how_it_works(self, video=False):
		"""
		Generates theory on how the algorithm works right in the Jupyter Notebook/Google colab.

		:param video: Some people prefer video tutorials over reading version. Set this parameter to True if you want video tutorial instead. :)
		"""
		if not super().__is_visual_on__():
			print("Supported only in Jupyter Notebook and Google Colab.")
			return NotImplementedError

		from IPython.core.getipython import get_ipython

		if not video:
			content = u"""
<div>
<h1>Machine Learning Basics with the K-Nearest Neighbors Algorithm</h1>
<br>
<br>

<p>

<h2>K-Nearest Neighbors</h2><br><br>
The KNN algorithm assumes that similar things exist in close proximity. In other words, similar things are near to each other.
<br><br>
<img src="https://miro.medium.com/max/672/1*wW8O-0xVQUFhBGexx2B6hg.png">

<br><br>
Notice in the image above that most of the time, similar data points are close to each other. The KNN algorithm hinges on this assumption being true enough for the algorithm to be useful. KNN captures the idea of similarity (sometimes called distance, proximity, or closeness) with some mathematics we might have learned in our childhood— calculating the distance between points on a graph.
<br><br>Note: An understanding of how we calculate the distance between points on a graph is necessary before moving on. If you are unfamiliar with or need a refresher on how this calculation is done, thoroughly read “Distance Between 2 Points” in its entirety, and come right back.
<br><br>There are other ways of calculating distance, and one way might be preferable depending on the problem we are solving. However, the straight-line distance (also called the Euclidean distance) is a popular and familiar choice.
<br><br>
<h2>The KNN Algorithm</h2> <br><br><br><br>
    1. Load the data <br><br>
    2. Initialize K to your chosen number of neighbors <br><br>
    3. For each example in the data <br><br>
        3.1 Calculate the distance between the query example and the current example from the data. <br><br>
        3.2 Add the distance and the index of the example to an ordered collection <br><br>
    4. Sort the ordered collection of distances and indices from smallest to largest (in ascending order) by the distances <br><br>
    5. Pick the first K entries from the sorted collection <br><br>
    6. Get the labels of the selected K entries <br><br>
    7. If regression, return the mean of the K labels <br><br>
    8. If classification, return the mode of the K labels <br><br>
<br><br>
<h2>Choosing the right value for K</h2> <br><br><br><br>
To select the K that’s right for your data, we run the KNN algorithm several times with different values of K and choose the K that reduces the number of errors we encounter while maintaining the algorithm’s ability to accurately make predictions when it’s given data it hasn’t seen before.
<br><br><br><br>Here are some things to keep in mind:
<br><br>    1. As we decrease the value of K to 1, our predictions become less stable. Just think for a minute, imagine K=1 and we have a query point surrounded by several reds and one green (I’m thinking about the top left corner of the colored plot above), but the green is the single nearest neighbor. Reasonably, we would think the query point is most likely red, but because K=1, KNN incorrectly predicts that the query point is green.
<br><br>    2. Inversely, as we increase the value of K, our predictions become more stable due to majority voting / averaging, and thus, more likely to make more accurate predictions (up to a certain point). Eventually, we begin to witness an increasing number of errors. It is at this point we know we have pushed the value of K too far.
<br><br>    3. In cases where we are taking a majority vote (e.g. picking the mode in a classification problem) among labels, we usually make K an odd number to have a tiebreaker.
<br><br><br><br><h2>Advantages</h2><br><br>
<br><br>    1. The algorithm is simple and easy to implement.
<br><br>    2. There’s no need to build a model, tune several parameters, or make additional assumptions.
<br><br>    3. The algorithm is versatile. It can be used for classification, regression, and search (as we will see in the next section).
<br><br><br><br><h2>Disadvantages</h2><br><br>
    1. The algorithm gets significantly slower as the number of examples and/or predictors/independent variables increase.
</p>



<h1>Author and source:</h1>
<h2>Author: <a target="_blank" href="https://twitter.com/onelharrison">Onel Harrison</a></h2>
<h2>To find more resources go to the source of the post: <a target="_blank" href="https://towardsdatascience.com/machine-learning-basics-with-the-k-nearest-neighbors-algorithm-6a6e71d01761">Towards data science post</a></h2>

</div>
"""
			get_ipython().run_cell_magic(u'html', u'', content)

		else:
			content = u"""
<div>
<h1> K-Nearest Neighbors - How it works? </h1>
<iframe width="560" height="315" src="https://www.youtube.com/embed/44jq6ano5n0" frameborder="0" allow="accelerometer; autoplay; encrypted-media; gyroscope; picture-in-picture" allowfullscreen></iframe>
</div>
"""
			get_ipython().run_cell_magic(u'html', u'', content)

	def interview_questions(self):
		"""
		Generates commonly asked interview questions about the algorithm in the Jupyter Notebook/Google colab.
		"""
		if not super().__is_visual_on__():
			print("Supported only in Jupyter Notebook and Google Colab.")
			return NotImplementedError

		from IPython.core.getipython import get_ipython

		content = u"""
<h1> K-Nearest Neighbors Interview Questions </h1>

<h2> 1. What is “K” in KNN algorithm? </h2>
<p>
K = Number of nearest neighbors you want to select to predict the class of a given item
</p>

<h2> 2. How do we decide the value of "K" in KNN algorithm? </h2>
<p>
If K is small, then results might not be reliable because noise will have a higher influence on the result. If K is large, then there will be a lot of processing which may adversely impact the performance of the algorithm. So, following is must be considered while choosing the value of K:

a. K should be the square root of n (number of data points in training dataset)
b. K should be odd so that there are no ties. If square root is even, then add or subtract 1 to it.
</p>
<h2> 3. Why is the odd value of “K” preferable in KNN algorithm? </h2>
<p>
K should be odd so that there are no ties in the voting. If square root of number of data points is even, then add or subtract 1 to it to make it odd.
</p>
<h2> 4. What is the difference between Euclidean Distance and Manhattan distance? What is the formula of Euclidean distance and Manhattan distance? </h2>
<p>
Both are used to find out the distance between two points. <br><br>
  <img src="https://4.bp.blogspot.com/-9iDGWtgrbh0/XErhbogDiBI/AAAAAAAABLE/tAhILG2rJ68Hs88XBSi5PP0Wkxi3F-U2ACLcBGAs/s1600/Euclidean_distance_and_Manhattan_distance.PNG">
<br><br><br>
Euclidean Distance and Manhattan Distance Formula <br><br>
<img src="https://4.bp.blogspot.com/-Tr6BrJ4mZNw/XErlXGm8xrI/AAAAAAAABLc/ZWSmXXOrQBIyKmeO4DdPvckpUVjDFHW7wCLcBGAs/s1600/Euclidean_distance_and_Manhattan_distance_formula.PNG">
<br><br>
</p>
<h2> 5. Why is KNN algorithm called Lazy Learner? </h2>
<p>
When it gets the training data, it does not learn and make a model, it just stores the data. It does not derive any discriminative function from the training data. It uses the training data when it actually needs to do some prediction. So, KNN does not immediately learn a model, but delays the learning, that is why it is called lazy learner. 
</p>
<h2>  6. Why should we not use KNN algorithm for large datasets? </h2>
<p>
KNN works well with smaller dataset because it is a lazy learner. It needs to store all the data and then makes decision only at run time. It needs to calculate the distance of a given point with all other points. So if dataset is large, there will be a lot of processing which may adversely impact the performance of the algorithm. 

KNN is also very sensitive to noise in the dataset. If the dataset is large, there are chances of noise in the dataset which adversely affect the performance of KNN algorithm.	
</p>
<h3>  The questions and answers taken from: [<a href="http://theprofessionalspoint.blogspot.com/2019/01/knn-algorithm-in-machine-learning.html">link</a>]</h3>

<h3>  Quiz like questions: [<a href="https://www.analyticsvidhya.com/blog/2017/09/30-questions-test-k-nearest-neighbors-algorithm/">link</a>] </h3>
"""
		get_ipython().run_cell_magic(u'html', u'', content)
